import json
import os
from typing import Optional

from guarddog.analyzer.metadata.typosquatting import TyposquatDetector
from guarddog.utils.config import TOP_PACKAGES_CACHE_LOCATION


class GoTyposquatDetector(TyposquatDetector):
    """Detector for typosquatting attacks for go modules. Checks for distance one Levenshtein,
    one-off character swaps, permutations around hyphens, and substrings.

    Attributes:
        popular_packages (set): set of top 500 most popular Go packages,
          as determined by count of references across top starred repositories
    """

    def _get_top_packages(self) -> set:
        top_packages_filename = "top_go_packages.json"

        resources_dir = TOP_PACKAGES_CACHE_LOCATION
        if resources_dir is None:
            resources_dir = os.path.abspath(
                os.path.join(os.path.dirname(__file__), "..", "resources")
            )

        top_packages_path = os.path.join(resources_dir, top_packages_filename)

        top_packages_information = None

        if top_packages_filename in os.listdir(resources_dir):
            with open(top_packages_path, "r") as top_packages_file:
                top_packages_information = json.load(top_packages_file)

        if top_packages_information is None:
            raise Exception(
                f"Could not retrieve top Go packages from {top_packages_path}")

        return set(top_packages_information)

    def detect(
        self,
        package_info,
        path: Optional[str] = None,
        name: Optional[str] = None,
        version: Optional[str] = None,
    ) -> tuple[bool, Optional[str]]:
        """
        Uses a Go package's name to determine the
        package is attempting a typosquatting attack

        Args:
            name (str): The name of the package,
                also known as the import path

        Returns:
            Tuple[bool, Optional[str]]: True if package is typosquatted,
               along with a message indicating the similar package name.
               False if not typosquatted and None
        """

        similar_package_names = self.get_typosquatted_package(name)
        if len(similar_package_names) > 0:
            return True, TyposquatDetector.MESSAGE_TEMPLATE % ", ".join(
                similar_package_names
            )
        return False, None

    def _get_confused_forms(self, package_name) -> list:
        """
        Gets confused terms for Go packages
        Confused terms are:
            - golang to go swaps (or vice versa)
            - the removal of go/golang terms
            - gitlab.com to github.com swaps (or vice versa)

        Args:
            package_name (str): name of the package

        Returns:
            list: list of confused terms
        """

        confused_forms = []

        if package_name.startswith("github.com/"):
            replaced = package_name.replace("github.com/", "gitlab.com/", 1)
            confused_forms.append(replaced)
        elif package_name.startswith("gitlab.com/"):
            replaced = package_name.replace("gitlab.com/", "github.com/", 1)
            confused_forms.append(replaced)

        terms = package_name.split("-")

        # Detect swaps like golang-package -> go-package
        for i in range(len(terms)):
            confused_term = None

            if "golang" in terms[i]:
                confused_term = terms[i].replace("golang", "go")
            elif "go" in terms[i]:
                confused_term = terms[i].replace("go", "golang")
            else:
                continue

            # Get form when replacing or removing go/golang term
            replaced_form = terms[:i] + [confused_term] + terms[i + 1:]
            removed_form = terms[:i] + terms[i + 1:]

            for form in (replaced_form, removed_form):
                confused_forms.append("-".join(form))

        return confused_forms


if __name__ == "__main__":
    # update top_npm_packages.json
    GoTyposquatDetector()._get_top_packages()



from typing import Type

from guarddog.analyzer.metadata import Detector
from guarddog.analyzer.metadata.go.typosquatting import GoTyposquatDetector

GO_METADATA_RULES = {}

classes: list[Type[Detector]] = [
    GoTyposquatDetector,
]

for detectorClass in classes:
    detectorInstance = detectorClass()  # type: ignore
    GO_METADATA_RULES[detectorInstance.get_name()] = detectorInstance



from guarddog.scanners.npm_package_scanner import NPMPackageScanner  # NOQA
from guarddog.scanners.pypi_package_scanner import PypiPackageScanner  # NOQA
from guarddog.scanners.go_package_scanner import GoModuleScanner  # NOQA



import logging
import os
import stat
import zipfile

import tarsafe  # type:ignore

log = logging.getLogger("guarddog")


def is_supported_archive(path: str) -> bool:
    """
    Decide whether a file contains a supported archive based on its
    file extension.

    Args:
        path (str): The local filesystem path to examine

    Returns:
        bool: Represents the decision reached for the file

    """
    def is_tar_archive(path: str) -> bool:
        tar_exts = [".bz2", ".bzip2", ".gz", ".gzip", ".tgz", ".xz"]

        return any(path.endswith(ext) for ext in tar_exts)

    def is_zip_archive(path: str) -> bool:
        return any(path.endswith(ext) for ext in [".zip", ".whl", ".egg"])

    return is_tar_archive(path) or is_zip_archive(path)


def safe_extract(source_archive: str, target_directory: str) -> None:
    """
    safe_extract safely extracts archives to a target directory.

    This function does not clean up the original archive and does not
    create the target directory if it does not exist.  It also assumes
    the source archive argument is a path to a regular file on the
    local filesystem.

    @param source_archive:      The archive to extract
    @param target_directory:    The directory where to extract the archive to
    @raise ValueError           If the archive type is unsupported

    """
    log.debug(f"Extracting archive {source_archive} to directory {target_directory}")
    if tarsafe.is_tarfile(source_archive):

        def add_exec(path):
            st = os.stat(path)
            os.chmod(path, st.st_mode | stat.S_IEXEC)

        def add_read(path):
            st = os.stat(path)
            os.chmod(path, st.st_mode | stat.S_IREAD)

        def recurse_add_perms(path):
            add_exec(path)
            for root, dirs, files in os.walk(path):
                for d in dirs:
                    add_exec(os.path.join(root, d))
                for f in files:
                    add_read(os.path.join(root, f))

        tarsafe.open(source_archive).extractall(target_directory)
        recurse_add_perms(target_directory)

    elif zipfile.is_zipfile(source_archive):
        with zipfile.ZipFile(source_archive, 'r') as zip:
            for file in zip.namelist():
                # Note: zip.extract cleans up any malicious file name
                # such as directory traversal attempts This is not the
                # case of zipfile.extractall
                zip.extract(file, path=os.path.join(target_directory, file))
    else:
        raise ValueError(f"unsupported archive extension: {source_archive}")



import os
import re
import pathlib
from dataclasses import dataclass
from typing import Optional, Iterable

import yaml
from yaml.loader import SafeLoader

from guarddog.ecosystems import ECOSYSTEM

current_dir = pathlib.Path(__file__).parent.resolve()


# These data class aim to reduce the spreading of the logic
# Instead of using the a dict as a structure and parse it difffently depending on the type
@dataclass
class SourceCodeRule:
    """
    Base class for source code rules
    """
    id: str
    file: str
    description: str


@dataclass
class YaraRule(SourceCodeRule):
    """
    Yara rule just reimplements base
    """
    pass


@dataclass
class SempgrepRule(SourceCodeRule):
    """
    Semgrep rule are language specific
    Content of rule in yaml format is accessible through rule_content
    """
    ecosystem: ECOSYSTEM
    rule_content: dict


def get_sourcecode_rules(
    ecosystem: ECOSYSTEM, kind: Optional[type] = None
) -> Iterable[SourceCodeRule]:
    """
    This function returns the source code rules for a given ecosystem and kind.
    Args:
        ecosystem: The ecosystem to filter for if rules are ecosystem specific
        kind: The kind of rule to filter for
    """
    for rule in SOURCECODE_RULES:
        if kind and not isinstance(rule, kind):
            continue
        if not (getattr(rule, "ecosystem", ecosystem) == ecosystem):
            continue
        yield rule


SOURCECODE_RULES: list[SourceCodeRule] = list()

semgrep_rule_file_names = list(
    filter(lambda x: x.endswith("yml"), os.listdir(current_dir))
)
# all yml files placed in the sourcecode directory are loaded as semgrep rules
# refer to README.md for more information
for file_name in semgrep_rule_file_names:
    with open(os.path.join(current_dir, file_name), "r") as fd:
        data = yaml.load(fd, Loader=SafeLoader)
        for rule in data["rules"]:
            for lang in rule["languages"]:
                ecosystem = None
                match lang:
                    case "python":
                        ecosystem = ECOSYSTEM.PYPI
                    case "javascript" | "typescript" | "json":
                        ecosystem = ECOSYSTEM.NPM
                    case "go":
                        ecosystem = ECOSYSTEM.GO
                    case _:
                        continue

                # avoids duplicates when multiple languages are supported by a rule
                if not next(
                    filter(
                        lambda r: r.id == rule["id"],
                        get_sourcecode_rules(ecosystem, SempgrepRule),
                    ),
                    None,
                ):
                    SOURCECODE_RULES.append(
                        SempgrepRule(
                            id=rule["id"],
                            ecosystem=ecosystem,
                            description=rule.get("metadata", {}).get("description", ""),
                            file=file_name,
                            rule_content=rule,
                        )
                    )

yara_rule_file_names = list(
    filter(lambda x: x.endswith("yar"), os.listdir(current_dir))
)
# all yar files placed in the sourcecode directory are loaded as YARA rules
# refer to README.md for more information
for file_name in yara_rule_file_names:
    rule_id = pathlib.Path(file_name).stem
    description_regex = fr'\s*rule\s+{rule_id}[^}}]+meta:[^}}]+description\s*=\s*\"(.+?)\"'

    with open(os.path.join(current_dir, file_name), "r") as fd:
        match = re.search(description_regex, fd.read())
        rule_description = ""
        if match:
            rule_description = match.group(1)
        SOURCECODE_RULES.append(YaraRule(id=rule_id, file=file_name, description=rule_description))



from typing import Optional

from guarddog.analyzer.metadata.unclaimed_maintainer_email_domain import \
    UnclaimedMaintainerEmailDomainDetector

from .utils import NPM_API_MAINTAINER_EMAIL_WARNING, get_email_addresses


class NPMUnclaimedMaintainerEmailDomainDetector(UnclaimedMaintainerEmailDomainDetector):
    def __init__(self):
        super().__init__("npm")

        self.description += "; " + NPM_API_MAINTAINER_EMAIL_WARNING

    def get_email_addresses(self, package_info: dict) -> set[str]:
        return get_email_addresses(package_info)

    def detect(
        self,
        package_info,
        path: Optional[str] = None,
        name: Optional[str] = None,
        version: Optional[str] = None,
    ) -> tuple[bool, str]:
        has_issues, message = super().detect(package_info, path, name, version)

        if has_issues:
            message += "\n\n" + NPM_API_MAINTAINER_EMAIL_WARNING

        return has_issues, message



from abc import abstractmethod
import hashlib
import logging
import os
from typing import Optional

from guarddog.analyzer.metadata.detector import Detector

log = logging.getLogger("guarddog")


class BundledBinary(Detector):
    """This heuristic detects if the package includes bundles binary file."""

    # magic bytes are the first few bytes of a file that can be used to identify the file type
    # regardless of their extension
    magic_bytes = {"exe": b"\x4D\x5A", "elf": b"\x7F\x45\x4C\x46"}

    def __init__(self):
        super().__init__(
            name="bundled_binary", description="Identify packages bundling binaries"
        )

    @abstractmethod
    def detect(
        self,
        package_info,
        path: Optional[str] = None,
        name: Optional[str] = None,
        version: Optional[str] = None,
    ) -> tuple[bool, str]:
        def format_file(file: str, kind: str) -> str:
            return f"{file} ({kind})"

        def sha256(file: str) -> str:
            with open(file, "rb") as f:
                hasher = hashlib.sha256()
                while (chunk := f.read(4096)):
                    hasher.update(chunk)
                return hasher.hexdigest()

        log.debug(
            f"Running bundled binary heuristic on package {name} version {version}"
        )
        if not path:
            raise ValueError("path is needed to run heuristic " + self.get_name())

        bin_files = {}
        for root, _, files in os.walk(path):
            for f in files:
                path = os.path.join(root, f)
                kind = self.is_binary(path)
                if kind:
                    digest = sha256(path)
                    if digest not in bin_files:
                        bin_files[digest] = [format_file(f, kind)]
                    else:
                        bin_files[digest].append(format_file(f, kind))

        if not bin_files:
            return False, ""

        output_lines = '\n'.join(
            f"{digest}: {', '.join(files)}" for digest, files in bin_files.items()
        )
        return True, f"Binary file/s detected in package:\n{output_lines}"

    def is_binary(self, path: str) -> Optional[str]:
        max_head = len(max(self.magic_bytes.values()))
        with open(os.path.join(path), "rb") as fd:
            header: bytes = fd.read(max_head)
            for k, v in self.magic_bytes.items():
                if header[: len(v)] == v:
                    return k
        return None



import logging
from typing import Tuple
import os
import requests

from guarddog.analyzer.analyzer import Analyzer
from guarddog.ecosystems import ECOSYSTEM
from guarddog.scanners.scanner import PackageScanner

log = logging.getLogger("guarddog")

# See https://go.dev/ref/mod#goproxy-protocol to learn more about the Go modules proxy internals

# TODO: allow users to configure the proxy they wanna use.
GOPROXY_URL = "https://proxy.golang.org"


class GoModuleScanner(PackageScanner):
    def __init__(self) -> None:
        super().__init__(Analyzer(ECOSYSTEM.GO))

    def download_and_get_package_info(
        self, directory: str, package_name: str, version=None
    ) -> Tuple[dict, str]:
        # If the version is not set explicitely, guarddog defaults to the latest
        if not version:
            latest_version_info_url = (
                f"{GOPROXY_URL}/{escape_module_name(package_name)}/@latest"
            )
            log.debug(
                f"Version for Go module {package_name} is unspecified, "
                f"fetching the latest version info from {latest_version_info_url}..."
            )
            latest_version_info = requests.get(latest_version_info_url)
            latest_version_info.raise_for_status()
            latest_version = latest_version_info.json()["Version"]
            log.debug(
                f"Latest version for Go module {package_name} is {latest_version}"
            )
            version = latest_version

        # Most of this logic comes from the NPM package scanner
        zip_url = f"{GOPROXY_URL}/{escape_module_name(package_name)}/@v/{version}.zip"
        zip_path = os.path.join(directory, package_name.replace("/", "-") + ".zip")
        unzipped_path = zip_path.removesuffix(".zip")
        self.download_compressed(zip_url, zip_path, unzipped_path)

        version_info_url = (
            f"{GOPROXY_URL}/{escape_module_name(package_name)}/@v/{version}.info"
        )
        log.debug(
            f"Fetching Go module {package_name}@{version}'s info from {version_info_url}..."
        )
        version_info = requests.get(version_info_url)
        version_info.raise_for_status()

        return version_info.json(), unzipped_path


# As described in https://go.dev/ref/mod#goproxy-protocol:
# > To avoid ambiguity when serving from case-insensitive file systems,
# > the $module and $version elements are case-encoded by replacing every uppercase letter
# > with an exclamation mark followed by the corresponding lower-case letter.
# > This allows modules example.com/M and example.com/m to both be stored on disk,
# > since the former is encoded as example.com/!m.
def escape_module_name(package_name: str) -> str:
    escaped_package_name = ""

    for c in package_name:
        if c.isupper():
            escaped_package_name += f"!{c.lower()}"
        else:
            escaped_package_name += c

    return escaped_package_name



import os
import multiprocessing

"""
This parameter specifies the amount of threads to use while requiring parallelism
- Default: Number of CPUs available
"""
PARALLELISM: int = int(
    os.environ.get("GUARDDOG_PARALLELISM", multiprocessing.cpu_count())
)

"""
This flag specifies if an analysis of all posible versions is required
- True: All possible versions are analyzed
- False [default]: Only best match is analyzed
"""
VERIFY_EXHAUSTIVE_DEPENDENCIES: bool = (
    os.environ.get("GUARDDOG_VERIFY_EXHAUSTIVE_DEPENDENCIES", "false").lower() == "true"
)

"""
This parameter specifies the location of the top packages cache
- Default: guarddog/analyzer/metadata/resources
"""
TOP_PACKAGES_CACHE_LOCATION: str = os.environ.get(
    "GUARDDOG_TOP_PACKAGES_CACHE_LOCATION",
    os.path.abspath(
        os.path.join(os.path.dirname(__file__), "../analyzer/metadata/resources")
    ),
)

"""
This parameter specifies comman separated file extentions that YARA rules will not run against
- Default: ini,md,rst,txt,lock,json,yaml,yml,toml,xml,html,rst,csv,sql,pdf,doc,docx,ppt,
           pptx,xls,xlsx,odt,changelog,readme,makefile,dockerfile,pkg-info
"""
YARA_EXT_EXCLUDE: list[str] = os.environ.get(
    "GUARDDOG_YARA_EXT_EXCLUDE",
    "ini,md,rst,txt,lock,json,yaml,yml,toml,xml,html,rst,csv,sql,pdf,doc,docx,ppt,"
    "pptx,xls,xlsx,odt,changelog,readme,makefile,dockerfile,pkg-info",
).split(",")



from abc import abstractmethod
from typing import Optional
import os

from disposable_email_domains import blocklist
from guarddog.analyzer.metadata.detector import Detector

from .utils import extract_email_address_domain


class DeceptiveAuthorDetector(Detector):
    """This heuristic detects when an author is using a disposable email."""

    MESSAGE_TEMPLATE = "An author was detected using a disposable email %s"

    def __init__(self):
        super().__init__(
            name="deceptive_author",
            description="This heuristic detects when an author is using a disposable email",
        )

    def get_email_addresses(self, package_info: dict) -> set[str]:
        return set()

    @staticmethod
    def get_suspicious_email_domains() -> set:
        """
        Gets the domains that are known to be used by suspicious authors.
        """
        # Obtain the path to the file containing knonw placeholder email domains
        placeholder_email_domains_filename = "placeholder_email_domains.txt"

        resources_dir = os.path.abspath(
            os.path.join(os.path.dirname(__file__), "resources")
        )

        placeholder_email_domains_path = os.path.join(
            resources_dir, placeholder_email_domains_filename
        )

        # read internal maintained list of placeholder email domains
        # this domains are usually used by authors who want to don't want to reveal their real email
        placeholder_email_domains_data = None
        with open(placeholder_email_domains_path, "r") as placeholder_email_domains_file:
            placeholder_email_domains_data = set(
                placeholder_email_domains_file.read().split("\n")
            )

        return blocklist | placeholder_email_domains_data

    def _get_normalized_domain(self, domain: str) -> str:
        """
        Returns a normalized version of the domain address.
        The format should be ready to compare against the list of known disposable domains.
        """
        domain = domain.lower().strip()
        return ".".join(domain.split(".")[-2:])

    @abstractmethod
    def detect(
        self,
        package_info,
        path: Optional[str] = None,
        name: Optional[str] = None,
        version: Optional[str] = None,
    ) -> tuple[bool, str]:
        emails = self.get_email_addresses(package_info)

        if len(emails) == 0:
            # No e-mail is set for this package, hence no risk
            return False, "No e-mail found for this package"

        for email in emails:
            domain = self._get_normalized_domain(extract_email_address_domain(email))

            if domain in DeceptiveAuthorDetector.get_suspicious_email_domains():
                return True, self.MESSAGE_TEMPLATE % email
        return False, ""



""" Compromised Email Detector

Detects if a maintainer's email domain might have been compromised.
"""

from datetime import datetime, timezone
from typing import Optional

from dateutil import parser

from guarddog.analyzer.metadata.potentially_compromised_email_domain import \
    PotentiallyCompromisedEmailDomainDetector

from .utils import NPM_API_MAINTAINER_EMAIL_WARNING, get_email_addresses


class NPMPotentiallyCompromisedEmailDomainDetector(
    PotentiallyCompromisedEmailDomainDetector
):
    def __init__(self):
        super().__init__("npm")

        self.description += "; " + NPM_API_MAINTAINER_EMAIL_WARNING

    def get_email_addresses(self, package_info: dict) -> set[str]:
        return get_email_addresses(package_info)

    def detect(
        self,
        package_info,
        path: Optional[str] = None,
        name: Optional[str] = None,
        version: Optional[str] = None,
    ) -> tuple[bool, str]:
        has_issues, message = super().detect(package_info, path, name, version)

        if has_issues:
            message += "\n\n" + NPM_API_MAINTAINER_EMAIL_WARNING

        return has_issues, message

    def get_project_latest_release_date(self, package_info) -> Optional[datetime]:
        """
        Gets the most recent release date of a Python project

        Args:
            releases (dict): PyPI JSON API's representation field

        Returns:
            datetime: creation date of the most recent in releases
        """
        latest_release_version = package_info["dist-tags"]["latest"]
        raw_date = package_info["time"][latest_release_version]
        release_date = parser.isoparse(raw_date).replace(tzinfo=timezone.utc)
        return release_date
